/* jshint node:true */

'use strict';

var superagent = require('superagent'),
    util = require('util'),
    path = require('path'),
    assert = require('assert'),
    fs = require('fs'),
    os = require('os'),
    safe = require('safetydance'),
    Table = require('easy-table'),
    readlineSync = require('readline-sync'),
    config = require('../config.js'),
    tar = require('tar-fs'),
    helper = require('../helper.js'),
    execSync = require('child_process').execSync,
    exit = helper.exit,
    EventSource = require('eventsource'),
    manifestFormat = require('cloudron-manifestformat'),
    semver = require('semver'),
    split = require('split'),
    micromatch = require('micromatch');

require('colors');

exports = module.exports = {
    login: login,
    logout: logout,
    info: info,
    listVersions: listVersions,
    submit: submit,
    upload: upload,
    revoke: revoke,
    unpublish: unpublish,
    build: build,
    buildLogs: buildLogs,
    listPublishedApps: listPublishedApps
};

const NO_MANIFEST_FOUND_ERROR_STRING = 'No CloudronManifest.json found';

function createUrl(api) {
    return config.appStoreOrigin() + api;
}

// takes a function returning a superagent request instance and will reauthenticate in case the token is invalid
function superagentEnd(requestFactory, callback) {
    requestFactory().end(function (error, result) {
        if (error && !error.response) return callback(error);
        if (result.statusCode === 401) return authenticate({ error: true }, superagentEnd.bind(null, requestFactory, callback));
        if (result.statusCode === 403) return callback(new Error(result.type === 'application/javascript' ? JSON.stringify(result.body) : result.text));
        callback(error, result);
    });
}

function authenticate(options, callback) {
    if (!options.hideBanner) console.log('cloudron.io login'.bold + ' (If you do not have one, sign up at https://cloudron.io/console.html#/register)');

    var email = options.email || readlineSync.question('Email: ', {});
    var password = options.password || readlineSync.question('Password: ', { noEchoBack: true });

    config.set('appStoreToken', null);

    superagent.post(createUrl('/api/v1/login')).auth(email, password).send({ totpToken: options.totpToken }).end(function (error, result) {
        if (error && !error.response) exit(error);

        if (result.statusCode === 401 && result.body.message.indexOf('TOTP') !== -1) {
            if (result.body.message === 'TOTP token missing') console.log('A 2FA TOTP Token is required for this account.'.red);

            options.totpToken = readlineSync.question('2FA token: ', {});
            options.email = email;
            options.password = password;
            options.hideBanner = true;

            return authenticate(options, callback);
        }

        if (result.statusCode !== 200) {
            console.log('Login failed.'.red);

            options.hideBanner = true;
            options.email = '';
            options.password = '';

            return authenticate(options, callback);
        }

        config.set('appStoreToken', result.body.accessToken);

        console.log('Login successful.'.green);

        if (typeof callback === 'function') callback();
    });
}

function login(options) {
    authenticate(options);
}

function logout() {
    config.set('appStoreToken', null);
    console.log('Done.'.green);
}

function info(options) {
    helper.getAppstoreId(options.appstoreId, function (error, id, version) {
        if (error) exit(error);

        superagentEnd(function () {
            return superagent.get(createUrl('/api/v1/developers/apps/' + id + '/versions/' + version)).query({ accessToken: config.appStoreToken() });
        }, function (error, result) {
            if (error && !error.response) exit(util.format('Failed to list apps: %s', error.message.red));
            if (result.statusCode !== 200) exit(util.format('Failed to list apps: %s message: %s', result.statusCode, result.text));

            var manifest = result.body.manifest;
            console.log('id: %s', manifest.id.bold);
            console.log('title: %s', manifest.title.bold);
            console.log('tagline: %s', manifest.tagline.bold);
            console.log('description: %s', manifest.description.bold);
            console.log('website: %s', manifest.website.bold);
            console.log('contactEmail: %s', manifest.contactEmail.bold);
        });
    });
}

function listVersions(options) {
    helper.verifyArguments(arguments);

    helper.getAppstoreId(options.appstoreId, function (error, id) {
        if (error) exit(error);

        superagentEnd(function () {
            return superagent.get(createUrl('/api/v1/developers/apps/' + id + '/versions')).query({ accessToken: config.appStoreToken() });
        }, function (error, result) {
            if (error && !error.response) exit(util.format('Failed to list versions: %s', error.message.red));
            if (result.statusCode === 404) exit('This app is not listed in appstore');
            if (result.statusCode !== 200) exit(util.format('Failed to list versions: %s message: %s', result.statusCode, result.text));

            if (result.body.versions.length === 0) return console.log('No versions found.');

            if (options.raw) return console.log(JSON.stringify(result.body.versions, null, 2));

            var versions = result.body.versions.reverse();

            // var manifest = versions[0].manifest;
            var t = new Table();

            versions.forEach(function (version) {
                t.cell('Version', version.manifest.version);
                t.cell('Creation Date', version.creationDate);
                t.cell('Publish state', version.publishState);
                t.newRow();
            });

            console.log();
            console.log(t.toString());
        });
    });
}

function addApp(manifest, baseDir, callback) {
    assert(typeof manifest === 'object');
    assert(typeof baseDir === 'string');
    assert(typeof callback === 'function');

    superagentEnd(function () {
        return superagent.post(createUrl('/api/v1/developers/apps'))
            .query({ accessToken: config.appStoreToken() })
            .send({ id: manifest.id });
    }, function (error, result) {
        if (error && !error.response) return exit(util.format('Failed to create app: %s', error.message.red));
        if (result.statusCode !== 201 && result.statusCode !== 409) {
            return exit(util.format('Failed to create app: %s message: %s', result.statusCode, result.text));
        }

        if (result.statusCode === 201) {
            console.log('New application added to the appstore with id %s.'.green, manifest.id);
        }

        callback();
    });
}

function parseChangelog(file, version) {
    var changelog = '';
    var data = safe.fs.readFileSync(file, 'utf8');
    if (!data) return null;
    var lines = data.split('\n');

    version = version.replace(/-.*/, ''); // remove any prerelease

    for (var i = 0; i < lines.length; i++) {
        if (lines[i] === '[' + version + ']') break;
    }

    for (i = i + 1; i < lines.length; i++) {
        if (lines[i] === '') continue;
        if (lines[i][0] === '[') break;

        changelog += lines[i] + '\n';
    }

    return changelog;
}

function addVersion(manifest, buildId, baseDir, callback) {
    assert(typeof manifest === 'object');
    assert(typeof buildId === 'string');
    assert(typeof baseDir === 'string');

    var iconFilePath = null;
    if (manifest.icon) {
        var iconFile = manifest.icon; // backward compat
        if (iconFile.slice(0, 7) === 'file://') iconFile = iconFile.slice(7);

        iconFilePath = path.isAbsolute(iconFile) ? iconFile : path.join(baseDir, iconFile);
        if (!fs.existsSync(iconFilePath)) return callback(new Error('icon not found at ' + iconFilePath));
    }

    if (manifest.description.slice(0, 7) === 'file://') {
        var descriptionFilePath = manifest.description.slice(7);
        descriptionFilePath = path.isAbsolute(descriptionFilePath) ? descriptionFilePath : path.join(baseDir, descriptionFilePath);
        manifest.description = safe.fs.readFileSync(descriptionFilePath, 'utf8');
        if (!manifest.description && safe.error) return callback(new Error('Could not read/parse description ' + safe.error.message));
        if (!manifest.description) return callback(new Error('Description cannot be empty'));
    }

    if (manifest.postInstallMessage && manifest.postInstallMessage.slice(0, 7) === 'file://') {
        var postInstallFilePath = manifest.postInstallMessage.slice(7);
        postInstallFilePath = path.isAbsolute(postInstallFilePath) ? postInstallFilePath : path.join(baseDir, postInstallFilePath);
        manifest.postInstallMessage = safe.fs.readFileSync(postInstallFilePath, 'utf8');
        if (!manifest.postInstallMessage && safe.error) return callback(new Error('Could not read/parse postInstall ' + safe.error.message));
        if (!manifest.postInstallMessage) return callback(new Error('PostInstall file specified but it is empty'));
    }

    if (manifest.changelog.slice(0, 7) === 'file://') {
        var changelogPath = manifest.changelog.slice(7);
        changelogPath = path.isAbsolute(changelogPath) ? changelogPath : path.join(baseDir, changelogPath);
        manifest.changelog = parseChangelog(changelogPath, manifest.version);
        if (!manifest.changelog) return callback(new Error('Bad changelog format or missing changelog for this version'));
    }

    superagentEnd(function () {
        var req = superagent.post(createUrl('/api/v1/developers/apps/' + manifest.id + '/versions'));
        req.query({ accessToken: config.appStoreToken() });
        if (iconFilePath) req.attach('icon', iconFilePath);
        req.field('buildId', buildId);
        req.attach('manifest', new Buffer(JSON.stringify(manifest)), 'manifest');
        return req;
    }, function (error, result) {
        if (error && !error.response) return callback(new Error(util.format('Failed to publish version: %s', error.message)));
        if (result.statusCode === 409) return callback('This version already exists. Use --force to overwrite.');
        if (result.statusCode !== 204) return callback(new Error(util.format('Failed to publish version (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text)));

        callback();
    });
}

function updateVersion(manifest, buildId, baseDir, callback) {
    assert(typeof manifest === 'object');
    assert(typeof buildId === 'string');
    assert(typeof baseDir === 'string');

    var iconFilePath = null;
    if (manifest.icon) {
        var iconFile = manifest.icon; // backward compat
        if (iconFile.slice(0, 7) === 'file://') iconFile = iconFile.slice(7);

        iconFilePath = path.isAbsolute(iconFile) ? iconFile : path.join(baseDir, iconFile);
        if (!fs.existsSync(iconFilePath)) return callback(new Error('icon not found at ' + iconFilePath));
    }

    if (manifest.description.slice(0, 7) === 'file://') {
        var descriptionFilePath = manifest.description.slice(7);
        descriptionFilePath = path.isAbsolute(descriptionFilePath) ? descriptionFilePath : path.join(baseDir, descriptionFilePath);
        manifest.description = safe.fs.readFileSync(descriptionFilePath, 'utf8');
        if (!manifest.description) return callback(new Error('Could not read description ' + safe.error.message));
    }

    if (manifest.postInstallMessage && manifest.postInstallMessage.slice(0, 7) === 'file://') {
        var postInstallFilePath = manifest.postInstallMessage.slice(7);
        postInstallFilePath = path.isAbsolute(postInstallFilePath) ? postInstallFilePath : path.join(baseDir, postInstallFilePath);
        manifest.postInstallMessage = safe.fs.readFileSync(postInstallFilePath, 'utf8');
        if (!manifest.postInstallMessage) return callback(new Error('Could not read/parse postInstall ' + safe.error.message));
    }

    if (manifest.changelog.slice(0, 7) === 'file://') {
        var changelogPath = manifest.changelog.slice(7);
        changelogPath = path.isAbsolute(changelogPath) ? changelogPath : path.join(baseDir, changelogPath);
        manifest.changelog = parseChangelog(changelogPath, manifest.version);
        if (!manifest.changelog) return callback(new Error('Could not read changelog or missing version changes'));
    }

    superagentEnd(function () {
        var req = superagent.put(createUrl('/api/v1/developers/apps/' + manifest.id + '/versions/' + manifest.version));
        req.query({ accessToken: config.appStoreToken() });
        if (iconFilePath) req.attach('icon', iconFilePath);
        req.field('buildId', buildId);
        req.attach('manifest', new Buffer(JSON.stringify(manifest)), 'manifest');
        return req;
    }, function (error, result) {
        if (error && !error.response) return callback(new Error(util.format('Failed to publish version: %s', error.message)));
        if (result.statusCode !== 204) {
            return callback(new Error(util.format('Failed to publish version (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text)));
        }

        callback();
    });
}

function delVersion(manifest, force) {
    assert(typeof manifest === 'object');
    assert(typeof force === 'boolean');

    if (!force) {
        console.log('This will delete the version %s of app %s from the appstore!'.red, manifest.version.bold, manifest.id.bold);
        var reallyDelete = readlineSync.question(util.format('Really do this? [y/N]: '), {});
        if (reallyDelete.toUpperCase() !== 'Y') exit();
    }

    superagentEnd(function () {
        return superagent.del(createUrl('/api/v1/developers/apps/' + manifest.id + '/versions/' + manifest.version)).query({ accessToken: config.appStoreToken() });
    }, function (error, result) {
        if (error && !error.response) return exit(util.format('Failed to unpublish version: %s', error.message.red));
        if (result.statusCode !== 204) exit(util.format('Failed to unpublish version (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text));

        console.log('version unpublished.'.green);
    });
}

function revokeVersion(appstoreId, version) {
    assert.strictEqual(typeof appstoreId, 'string');
    assert.strictEqual(typeof version, 'string');

    superagentEnd(function () {
        return superagent.post(createUrl('/api/v1/developers/apps/' + appstoreId + '/versions/' + version + '/revoke'))
            .query({ accessToken: config.appStoreToken() })
            .send({ });
    }, function (error, result) {
        if (error && !error.response) return exit(util.format('Failed to revoke version: %s', error.message.red));
        if (result.statusCode !== 200) exit(util.format('Failed to revoke version (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text));

        console.log('version revoked.'.green);
    });
}

function delApp(appId, force) {
    assert(typeof appId === 'string');
    assert(typeof force === 'boolean');

    if (!force) {
        console.log('This will delete app %s from the appstore!'.red, appId.bold);
        var reallyDelete = readlineSync.question(util.format('Really do this? [y/N]: '), {});
        if (reallyDelete.toUpperCase() !== 'Y') exit();
    }

    superagentEnd(function () {
        return superagent.del(createUrl('/api/v1/developers/apps/' + appId)).query({ accessToken: config.appStoreToken() });
    }, function (error, result) {
        if (error && !error.response) return exit(util.format('Failed to unpublish app: %s', error.message.red));
        if (result.statusCode !== 204) exit(util.format('Failed to unpublish app (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text));

        console.log('App unpublished.'.green);
    });
}

function submitAppForReview(manifest, callback) {
    assert(typeof manifest === 'object');
    assert(typeof callback === 'function');

    superagentEnd(function () {
        return superagent.post(createUrl('/api/v1/developers/apps/' + manifest.id + '/versions/' + manifest.version + '/submit'))
            .query({ accessToken: config.appStoreToken() })
            .send({ });
    }, function (error, result) {
        if (error && !error.response) exit(util.format('Failed to submit app for review: %s', error.message.red));
        if (result.statusCode === 404) {
            console.log('No version %s found. Please use %s first.', manifest.version.bold, 'cloudron appstore upload'.cyan);
            exit('Failed to submit app for review.'.red);
        }
        if (result.statusCode !== 200) return exit(util.format('Failed to submit app (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text));

        console.log('App submitted for review.'.green);
        console.log('You will receive an email when approved.');

        callback();
    });
}

function upload(options) {
    helper.verifyArguments(arguments);

    // try to find the manifest of this project
    var manifestFilePath = helper.locateManifest();
    if (!manifestFilePath) return exit(NO_MANIFEST_FOUND_ERROR_STRING);

    var result = manifestFormat.parseFile(manifestFilePath);
    if (result.error) return exit(result.error.message);

    var error = manifestFormat.checkAppstoreRequirements(result.manifest);
    if (error) {
        if (!options.skipValidation) return exit(error.message);

        console.log(error.message.red);

        var reallyUpload = readlineSync.question(util.format('Appstore requirements are not met. Really upload? [y/N]: '), {});
        if (reallyUpload.toUpperCase() !== 'Y') exit();
    }

    var manifest = result.manifest;

    // ensure the app is known on the appstore side
    addApp(manifest, path.dirname(manifestFilePath), function () {
        helper.selectBuild(manifest.id, true /* latest */, function (error, build) {
            if (error || !build.dockerImage) exit('No build found, please run `cloudron build` first and test the new build on your Cloudron.');

            console.log('Publishing %s@%s for %s with build %s.', manifest.id, manifest.version, 'testing'.yellow, build.id);

            var func = options.force ? updateVersion : addVersion;

            func(manifest, build.id, path.dirname(manifestFilePath), function (error) {
                if (error) return exit(error);

                console.log('\nThe App Store view\'s %s tab in your cloudron will show the app.', 'testing'.yellow);

                console.log('\nApp can be tested on other cloudrons using the cli tool:\n\t\t%s', ('cloudron install --appstore-id ' + manifest.id + '@' + manifest.version).white);

                if (config.apiEndpoint()) {
                    var url = helper.createUrl('/#/appstore/' + manifest.id + '?version=' + manifest.version);
                    console.log('\nDirect link to the app on your Cloudron:\n\t\t%s\n', url.white);
                }
            });
        });
    });
}

function submit() {
    helper.verifyArguments(arguments);

    // try to find the manifest of this project
    var manifestFilePath = helper.locateManifest();
    if (!manifestFilePath) return exit(NO_MANIFEST_FOUND_ERROR_STRING);

    var result = manifestFormat.parseFile(manifestFilePath);
    if (result.error) return exit(result.error.message);

    var manifest = result.manifest;

    submitAppForReview(manifest, exit);
}

function unpublish(options) {
    helper.verifyArguments(arguments);

    helper.getAppstoreId(options.appstoreId, function (error, id, version) {
        if (error) exit(error);

        if (!version) {
            console.log('Unpublishing ' + options.app);
            delApp(options.app, !!options.force);
            return;
        }

        console.log('Unpublishing ' + id + '@' + version);
        delVersion(manifest, !!options.force);
    });
}

function revoke(options) {
    helper.verifyArguments(arguments);

    helper.getAppstoreId(options.appstoreId, function (error, id, version) {
        if (error) return exit(error);

        if (!version) return exit('--appstore-id must be of the format id@version');

        console.log('Revoking ' + id + '@' + version);
        revokeVersion(id, version);
    });
}

function getBuildInfo(buildId, callback) {
    assert(typeof buildId === 'string');
    assert(typeof callback === 'function');

    superagentEnd(function () {
        return superagent.get(createUrl('/api/v1/developers/builds/' + buildId)).query({ accessToken: config.appStoreToken() });
    }, function (error, result) {
        if (error && !error.response) return callback(error);
        if (result.statusCode !== 200) return callback(util.format('Failed to get build.'.red, result.statusCode, result.text));

        if (result.body.status === 'success') return callback(null, result.body);
        if (result.body.status === 'error') return callback(null, result.body);

        // if status is neither error nor success, we have to wait for the buildbot to update the status
        getBuildInfo(buildId, callback);
    });
}

function printBuildLog(buildId, callback) {
    assert(typeof buildId === 'string');
    assert(typeof callback === 'function');

    superagentEnd(function () {
        return superagent.get(createUrl('/api/v1/developers/builds/' + buildId + '/log')).query({ accessToken: config.appStoreToken() });
    }, function (error, result) {
        if (error && !error.response) return callback(error);
        if (result.statusCode === 420) return callback('No build logs yet. Try again later.');
        if (result.statusCode !== 200) return callback(util.format('Failed to get build log.'.red, result.statusCode, result.text));

        var stream = result.pipe(split());

        stream.on('data', function (line) {
            if (line === '') return;
            console.log(line); // intentionally raw without json decode
        });

        stream.on('end', function () {
            console.log();
            callback();
        });
    });
}

function followBuildLog(buildId, raw, callback) {
    assert(typeof buildId === 'string');
    assert(typeof raw === 'boolean');

    var es = new EventSource(createUrl('/api/v1/developers/builds/' + buildId + '/logstream?accessToken=' + config.appStoreToken()));
    var prevId = null, prevWasStatus = false;

    es.on('message', function (e) {
        if (raw) return console.dir(e);

        var data = safe.JSON.parse(e.data);

        if (!data) return; // this is a bug in docker or our build server

        if (data.status) { // push log
            if (data.id && data.id === prevId) {
                // the code below does not work os x if the line wraps, maybe we should clip the text to window size?
                process.stdout.clearLine();
                process.stdout.cursorTo(0);
            } else if (prevWasStatus) {
                process.stdout.write('\n');
            }

            process.stdout.write(data.status + (data.id ? ' ' + data.id : '') + (data.progress ? ' ' + data.progress : ''));

            prevId = data.id;
            prevWasStatus = true;

            return;
        }

        if (prevWasStatus === true) {
            process.stdout.write('\n');
            prevId = null;
            prevWasStatus = false;
        }

        if (data.stream) { // build log
            process.stdout.write(data.stream);
        } else if (data.message) {
            console.log(data.message.yellow.bold);
        } else if (typeof data.error === 'string') {
            console.log(data.error.red.bold);
        } else if (data.error) {
            console.error(data.error);
        }
    });
    es.on('error', function (error) {
        if (raw) console.dir(error);

        if (error && error.status === 204) { // build already finished
            console.log('Building already finished. Fetching full logs'.cyan);
            console.log();

            return printBuildLog(buildId, callback);
        }

        // We sometimes get { type: 'error' } from es module when the server closes the socket. not clear why
        if (error && !error.status && error.type === 'error') error = null;

        callback(error && error.status ? error : null); // eventsource module really needs to give us better errors
    });
}

function verifyDockerfile(dockerFilePath) {
    var contents = safe.fs.readFileSync(dockerFilePath, 'utf8');
    if (!contents) return safe.error;

    var lines = contents.split('\n');
    for (var i = 0; i < lines.length; i++) {
        var result = lines[i].match(/^\s*FROM\s+cloudron\/base:(\d+\.\d+\.\d+)(?:@sha256:.*)?/i);
        if (!result) continue;

        if (!semver.valid(result[1])) return new Error('Invalid base image version');

        return null;
    }

    return new Error('Base image must be cloudron/base:<version>. Please see the latest version at https://cloudron.io/references/baseimage.html');
}

function dockerignoreMatcher(dockerignorePath) {
    var patterns = [];

    if (fs.existsSync(dockerignorePath)) {
        patterns = fs.readFileSync(dockerignorePath, 'utf8').split('\n');

        patterns = patterns.filter(function (line) { return line[0] !== '#'; });
        patterns = patterns.map(function (line) {
            var l = line.trim();

            while (l[0] === '/') l = l.slice(1);
            while (l[l.length-1] === '/') l = l.slice(0, -1);

            return l;
        });
        patterns = patterns.filter(function (line) { return line.length !== 0; });
    }

    return function ignore(path) {
        return micromatch([ path ], patterns, { dot: true }).length == 1;
    };
}

function buildLocal(manifest, sourceDir, options) {
    let buildConfig = helper.getBuildConfig(manifest.id);
    if (!('serveraddress' in buildConfig)) return exit('private app builds must be first configured using "cloudron config build"'.red);

    const tag = manifest.version, buildId = String(Date.now());

    console.log('Building %s@%s', manifest.id.bold, manifest.version.bold);
    console.log();

    let dockerImage;
    if (buildConfig.serveraddress) {
        dockerImage = `${buildConfig.serveraddress}/${buildConfig.repository}:${tag}`;
    } else {
        dockerImage = `${buildConfig.repository}:${tag}`;
    }
    execSync(`docker build -t ${dockerImage} ${sourceDir}`, { stdio: 'inherit' });

    console.log();
    console.log(`Pushing ${dockerImage}`.bold);

    execSync(`docker push ${dockerImage}`, { stdio: 'inherit' });

    helper.addBuild(manifest.id, buildId);
    helper.updateBuild(manifest.id, buildId, dockerImage);
}

function buildService(manifest, sourceDir, options) {
    console.log('Building %s@%s', manifest.id.bold, manifest.version.bold);
    console.log();

    var sourceArchiveFilePath = path.join(os.tmpdir(), manifest.id + '.tar.gz');
    var dockerignoreFilePath = path.join(sourceDir, '.dockerignore');
    var ignoreMatcher = dockerignoreMatcher(dockerignoreFilePath);

    var stream = tar.pack(sourceDir, {
        ignore: function (name) {
            return ignoreMatcher(name.slice(sourceDir.length + 1)); // make name as relative path
        }
    }).pipe(fs.createWriteStream(sourceArchiveFilePath));

    stream.on('error', function (error) {
        exit('Failed to create application source archive: ' + error);
    });

    stream.on('finish', function () {
        superagentEnd(function () {
            return superagent.post(createUrl('/api/v1/developers/builds'))
                .query({ accessToken: config.appStoreToken(), noCache: !options.cache, dockerfile: 'Dockerfile' })
                .field('appId', manifest.id)
                .attach('sourceArchive', sourceArchiveFilePath);
        }, function (error, result) {
            if (error && !error.response) return exit(util.format('Failed to build app: %s', error.message.red));
            if (result.statusCode === 413) exit('Failed to build app. The app source is too large.\nPlease adjust your .dockerignore file to only include neccessary files.'.red);
            if (result.statusCode !== 201) exit(util.format('Failed to build app (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text));

            var buildId = result.body.id;

            helper.addBuild(manifest.id, buildId);

            console.log('Build scheduled with id %s', buildId.cyan);
            console.log('Waiting for build to begin, this may take a bit...');

            followBuildLog(buildId, !!options.raw, function (error) {
                if (error) return exit('Error reading build log:', error);

                getBuildInfo(buildId, function (error, build) {
                    if (error) return exit('Error getting build info:', error);

                    if (build.status === 'success') {
                        helper.updateBuild(manifest.id, buildId, build.dockerImage);
                        console.log('Success'.green);
                        exit();
                    } else if (build.status === 'error') {
                        exit('App could not be built due to errors above');
                    } else {
                        exit('Build has unknown status ' + build.status);
                    }
                });
            });
        });
    });
}

function build(options) {
    helper.verifyArguments(arguments);

    // try to find the manifest of this project
    var manifestFilePath = helper.locateManifest();
    if (!manifestFilePath) return exit(NO_MANIFEST_FOUND_ERROR_STRING);

    var result = manifestFormat.parseFile(manifestFilePath);
    if (result.error) return exit('Error in CloudronManifest.json: ' + result.error.message.red);

    let manifest = result.manifest;
    const sourceDir = path.dirname(manifestFilePath);

    // todo: move this to the buildbot
    var error = verifyDockerfile(path.join(sourceDir, '/Dockerfile'));
    if (error) return exit(error.message.red);

    if (manifest.private) {
        buildLocal(manifest, sourceDir, options);
    } else {
        buildService(manifest, sourceDir, options);
    }
}

function buildLogs(options) {
    helper.verifyArguments(arguments);

    // try to find the manifest of this project
    var manifestFilePath = helper.locateManifest();
    if (!manifestFilePath) return exit(NO_MANIFEST_FOUND_ERROR_STRING);

    var result = manifestFormat.parseFile(manifestFilePath);
    if (result.error) return exit(result.error.message);

    var manifest = result.manifest;

    helper.selectBuild(manifest.id, true /* latest */, function (error, build) {
        if (error) exit('No build found, use cloudron build to create one');

        console.log('Getting logs of %s', build.id);

        if (options.tail) return followBuildLog(build.id, exit);

        printBuildLog(build.id, exit);
    });
}

// TODO currently no pagination, only needed once we have users with more than 100 apps
function listPublishedApps(options) {
    helper.verifyArguments(arguments);

    superagentEnd(function () {
        return superagent.get(createUrl('/api/v1/developers/apps?per_page=100'))
            .query({ accessToken: config.appStoreToken() })
            .send({ });
    }, function (error, result) {
        if (error && !error.response) return exit(util.format('Failed to get list of published apps: %s', error.message.red));
        if (result.statusCode !== 200) return exit(util.format('Failed to get list of published apps (statusCode %s): \n%s', result.statusCode, result.body && result.body.message ? result.body.message.red : result.text));

        if (result.body.apps.length === 0) return console.log('No apps published.');

        var t = new Table();

        result.body.apps.forEach(function (app) {
            t.cell('Id', app.id);
            t.cell('Title', app.manifest.title);
            t.cell('Latest Version', app.manifest.version);
            t.cell('Publish State', app.publishState);
            t.cell('Creation Date', new Date(app.creationDate));
            if (options.image) t.cell('Image', app.manifest.dockerImage);
            t.newRow();
        });

        console.log();
        console.log(t.toString());
    });
}
